""" Provides commands used to initialize org websites. """

import click
import html
import isodate
import re
import shutil
import textwrap

from collections import defaultdict
from datetime import date, datetime, timedelta
from libres.modules.errors import InvalidEmailAddress, AlreadyReservedError
from onegov.core.cli import command_group, pass_group_context, abort
from onegov.form import FormCollection
from onegov.org.formats import DigirezDB
from onegov.org.models import Organisation
from onegov.reservation import ResourceCollection
from onegov.ticket import TicketCollection
from onegov.user import UserCollection, User
from sqlalchemy import create_engine
from uuid import uuid4


cli = command_group()


@cli.command(context_settings={'creates_path': True})
@click.argument('name')
@click.option('--locale',
              default='de_CH', type=click.Choice(['de_CH', 'fr_CH']))
@pass_group_context
def add(group_context, name, locale):
    """ Adds an org with the given name to the database. For example:

        onegov-org --select '/onegov_org/evilcorp' add "Evilcorp"

    """

    def add_org(request, app):

        if app.session().query(Organisation).first():
            abort("{} already contains an organisation".format(
                group_context.selector))

        app.settings.org.create_new_organisation(app, name, locale=locale)

        click.echo("{} was created successfully".format(name))

    return add_org


@cli.command()
@pass_group_context
def delete(group_context):
    """ Deletes a single organisation matching the selector.

    Selectors matching multiple organisations are disabled for saftey reasons.

    """

    def delete_org(request, app):

        org = app.org.title
        confirmation = "Do you really want to DELETE {}?".format(org)

        if not click.confirm(confirmation):
            abort("Deletion process aborted")

        if app.has_filestorage:
            click.echo("Removing File Storage")

            for item in app.filestorage.listdir('.'):
                if app.filestorage.isdir(item):
                    app.filestorage.removedir(item, recursive=True, force=True)
                else:
                    app.filestorage.remove(item)

        if app.depot_storage_path and app.bound_storage_path:
            click.echo("Removing Depot Storage")
            shutil.rmtree(str(app.bound_storage_path.absolute()))

        if app.has_database_connection:
            click.echo("Dropping Database Schema")

            assert app.session_manager.is_valid_schema(app.schema)

            dsn = app.session_manager.dsn
            app.session_manager.session().close_all()
            app.session_manager.dispose()

            engine = create_engine(dsn)
            engine.execute('DROP SCHEMA "{}" CASCADE'.format(app.schema))
            engine.raw_connection().invalidate()
            engine.dispose()

        click.echo("{} was deleted successfully".format(org))

    return delete_org


@cli.command(name='import-digirez', context_settings={'singular': True})
@click.argument('accessdb', type=click.Path(exists=True, resolve_path=True))
@click.option('--min-date', default=None,
              help="Min date in the form '2016-12-31'")
@click.option('--ignore-booking-conflicts', default=False, is_flag=True,
              help="Ignore booking conflicts (TESTING ONlY)")
def import_digirez(accessdb, min_date, ignore_booking_conflicts):
    """ Imports a Digirez reservation database into onegov.org.

    Example:

        onegov-org --select '/orgs/govikon' import-digirez room_booking.mdb

    """

    if not shutil.which('mdb-export'):
        abort("Could not find 'mdb-export', please install mdbtools!")

    min_date = min_date and isodate.parse_date(min_date) or date.today()
    min_date = datetime(min_date.year, min_date.month, min_date.day)

    db = DigirezDB(accessdb)
    db.open()

    # XXX this is currently specific to what Gemeinde Rüti has, we could put
    # this into an external file however.
    form_definition = textwrap.dedent("""
        Name *= ___
        Telefon =___
        Zweck *= ___
        Einrichtungen =
            [ ] Konzertbestuhlung
            [ ] Bankettbestuhlung
            [ ] Flipchart
            [ ] Hellraumprojektor
            [ ] Kinoleinwand (Saal)
            [ ] Mobile Leinwand
            [ ] Lautsprecheranlage
            [ ] Podestelemente
            [ ] Flügel (Saal)
            [ ] Getränke
            [ ] Küche (Saal)
            [ ] Office (Keller)
        Bemerkungen = ...
    """)

    floor_hours = {
        'Amthaus': (8, 24),
        'Löwen': (0, 24),
        'Gemeinschaftszentrum': (8, 24)
    }

    title_map = {
        'Jugend- raum': 'Jugendraum',
        'Spielgruppen- Raum': 'Spielgruppenraum',
        'Vereins- raum': 'Vereinsraum'
    }

    def get_formdata(member, booking):

        einrichtungen = []

        if booking.kbest == '1':
            einrichtungen.append('Konzertbestuhlung')

        if booking.bbest == '1':
            einrichtungen.append('Bankettbestuhlung')

        if booking.flip == '1':
            einrichtungen.append('Flipchart')

        if booking.ohproj == '1':
            einrichtungen.append('Hellraumprojektor')

        if booking.klw == '1':
            einrichtungen.append('Kinoleinwand (Saal)')

        if booking.mlw == '1':
            einrichtungen.append('Mobile Leinwand')

        if booking.lspr == '1':
            einrichtungen.append('Lautsprecheranlage')

        if booking.pod == '1':
            einrichtungen.append('Podestelemente')

        if booking.flg == '1':
            einrichtungen.append('Flügel (Saal)')

        if booking.getr == '1':
            einrichtungen.append('Getränke')

        if booking.getr == '1':
            einrichtungen.append('Küche (Saal)')

        if booking.ofk == '1':
            einrichtungen.append('Office (Keller)')

        return {
            'name': booking.anspname or ' '.join((
                member.member_last_name,
                member.member_first_name
            )),
            'telefon': booking.ansptel or member.member_phone,
            'zweck': booking.title,
            'bemerkungen': booking.description,
            'einrichtungen': einrichtungen
        }

    def unescape_dictionary(d):
        for key, value in d.items():
            if isinstance(value, str):
                d[key] = html.unescape(value)
            if isinstance(value, dict):
                d[key] = unescape_dictionary(value)
            if isinstance(value, list):
                d[key] = [html.unescape(v) for v in value]

        return d

    email_fixes = (
        (re.compile(r'2da-capo-rueti.ch$'), '@da-capo.rueti.ch'),
        (re.compile(r'@mzol$'), '@mzol.ch'),
        (re.compile(r'^jo$'), 'info@rueti.ch'),
        (re.compile(r'@pro-vitalis$'), '@pro-vitalis.ch')
    )

    def run_import(request, app):

        # create all resources first, fails if at least one exists already
        print("Creating resources")

        resources = ResourceCollection(app.libres_context)
        floors = {f.id: f.floor_name for f in db.records.floors}
        resources_by_room = {}

        for room in db.records.room:
            resource = resources.add(
                title=title_map.get(room.room_name, room.room_name),
                timezone='Europe/Zurich',
                type='room',
                group=floors[room.floor_id],
                definition=form_definition,
                meta={'is_hidden_from_public': True}
            )

            resources_by_room[room.id] = resource

        # gather all information needed to create the allocations/reservations
        relevant_bookings = (
            b for b in db.records.room_booking
            if isodate.parse_datetime(b.hour_end) > min_date
        )

        # group by room id in addition to the multi_id, as digirez supports
        # reservations which span multiple rooms
        reservations = defaultdict(list)

        for booking in relevant_bookings:
            group = '-'.join((booking.room_id, booking.multi_id))
            reservations[group].append(booking)

        # get the max_date per resource
        max_dates = {}

        for reservation_group in reservations:
            resource_id = resources_by_room[reservation_group.split('-')[0]].id

            if resource_id not in max_dates:
                max_dates[resource_id] = min_date

            for booking in reservations[reservation_group]:
                date = isodate.parse_datetime(booking.hour_end)

                if max_dates[resource_id] < date:
                    max_dates[resource_id] = date

        # get the user associated with the migration (the first admin)
        users = UserCollection(app.session())
        user = users.query().filter(User.role == 'admin').first()

        if not user:
            abort('No admin user found')

        # get the member mails
        members = {}

        for member in db.records.members:
            members[member.member_id] = member

        # create an allocation for all days between min/max date
        print("Creating allocations")

        for resource in resources_by_room.values():

            # true if the resource does not have any bookings
            if resource.id not in max_dates:
                continue

            start_hour, end_hour = floor_hours[resource.group]
            days = (max_dates[resource.id].date() - min_date.date()).days

            first_day_start = datetime(
                min_date.year, min_date.month, min_date.day,
                start_hour
            )

            first_day_end = datetime(
                min_date.year, min_date.month, min_date.day,
                end_hour - 1, 59, 59, 999999
            )

            whole_day = (start_hour, end_hour) == (0, 24)

            for day in range(0, days + 1):
                start = first_day_start + timedelta(days=day)
                end = first_day_end + timedelta(days=day)

                resource.scheduler.allocate(
                    dates=(start, end),
                    partly_available=True,
                    whole_day=whole_day,
                )

        # create the reservations
        print("Creating reservations")

        booking_conflicts = 0

        for reservation_group, bookings in reservations.items():
            resource = resources_by_room[reservation_group.split('-')[0]]
            scheduler = resource.scheduler
            found_conflict = False

            session_id = uuid4()

            for booking in bookings:
                assert bookings[0].anspmail == booking.anspmail
                assert bookings[0].member_id == booking.member_id
                assert bookings[0].anspname == booking.anspname
                assert bookings[0].ansptel == booking.ansptel
                assert bookings[0].title == booking.title
                assert bookings[0].description == booking.description

                email = booking.anspmail
                email = email or members[booking.member_id].member_email

                for expression, replacement in email_fixes:
                    email = expression.sub(replacement, email)

                try:
                    token = scheduler.reserve(
                        email=email,
                        dates=(
                            isodate.parse_datetime(booking.hour_start),
                            isodate.parse_datetime(booking.hour_end)
                        ),
                        session_id=session_id,
                        single_token_per_session=True,
                        data={'accepted': True}  # accepted through ticket
                    )
                    token = token.hex
                except InvalidEmailAddress:
                    abort("{} is an invalid e-mail address".format(email))
                except AlreadyReservedError:
                    booking_conflicts += 1
                    found_conflict = True

                    print("Booking conflict in {} at {}".format(
                        resource.title, booking.hour_start
                    ))
                    pass

            if found_conflict:
                continue

            forms = FormCollection(app.session())
            form_data = get_formdata(members[booking.member_id], booking)
            form_data = unescape_dictionary(form_data)
            form = resource.form_class(data=form_data)

            if not form.validate():
                abort("{} failed the form check with {}".format(
                    form_data, form.errors
                ))

            submission = forms.submissions.add_external(
                form=form,
                state='pending',
                id=token
            )

            scheduler.queries.confirm_reservations_for_session(session_id)
            scheduler.approve_reservations(token)

            forms.submissions.complete_submission(submission)

            with forms.session.no_autoflush:
                ticket = TicketCollection(request.session).open_ticket(
                    handler_code='RSV', handler_id=token
                )
                ticket.accept_ticket(user)
                ticket.close_ticket()

        if not ignore_booking_conflicts and booking_conflicts:
            abort("There were {} booking conflicts, aborting".format(
                booking_conflicts
            ))

    return run_import



import os

from collections import OrderedDict
from datetime import datetime, date
from freezegun import freeze_time
from onegov.core.request import CoreRequest
from onegov.core.utils import module_path, rchop
from onegov.org.models import Clipboard, ImageFileCollection
from onegov.org.models import SiteCollection
from onegov.org.models.file import GroupFilesByDateMixin
from onegov.org.models.resource import SharedMethods
from onegov.page import PageCollection
from onegov_testing.utils import create_image
from pytz import utc


def test_clipboard(org_app):

    request = CoreRequest(environ={
        'PATH_INFO': '/',
        'SERVER_NAME': '',
        'SERVER_PORT': '',
        'SERVER_PROTOCOL': 'https'
    }, app=org_app)

    clipboard = Clipboard.from_url(request, 'https://google.ch')
    assert clipboard.url == 'https://google.ch'

    clipboard.store_in_session()
    assert clipboard.from_session(clipboard.request).url == 'https://google.ch'

    clipboard.clear()
    assert clipboard.from_session(clipboard.request).url is None

    clipboard = Clipboard(request, clipboard.token + 'x')
    assert clipboard.url is None


def test_news_years(session):

    collection = PageCollection(session)
    news = collection.add_root("News", type='news')
    one = collection.add(news, title="One", type='news')
    two = collection.add(news, title="Two", type='news')

    assert news.years == [datetime.utcnow().year]

    one.created = datetime(2016, 2, 1, tzinfo=utc)
    two.created = datetime(2015, 2, 1, tzinfo=utc)

    assert news.years == [2016, 2015]


def test_group_intervals():
    mixin = GroupFilesByDateMixin()

    intervals = list(mixin.get_date_intervals(datetime(2016, 1, 1)))

    assert intervals[0].name == 'This month'
    assert intervals[0].start.date() == date(2016, 1, 1)
    assert intervals[0].end.date() == date(2016, 1, 31)

    assert intervals[1].name == 'Last month'
    assert intervals[1].start.date() == date(2015, 12, 1)
    assert intervals[1].end.date() == date(2015, 12, 31)

    assert intervals[2].name == 'Last year'
    assert intervals[2].start.date() == date(2015, 1, 1)
    assert intervals[2].end.date() == date(2015, 11, 30)

    assert intervals[3].name == 'Older'
    assert intervals[3].start.date() == date(2000, 1, 1)
    assert intervals[3].end.date() == date(2014, 12, 31)

    intervals = list(mixin.get_date_intervals(datetime(2016, 2, 1)))

    assert intervals[0].name == 'This month'
    assert intervals[0].start.date() == date(2016, 2, 1)
    assert intervals[0].end.date() == date(2016, 2, 29)

    assert intervals[1].name == 'Last month'
    assert intervals[1].start.date() == date(2016, 1, 1)
    assert intervals[1].end.date() == date(2016, 1, 31)

    assert intervals[2].name == 'Last year'
    assert intervals[2].start.date() == date(2015, 1, 1)
    assert intervals[2].end.date() == date(2015, 12, 31)

    assert intervals[3].name == 'Older'
    assert intervals[3].start.date() == date(2000, 1, 1)
    assert intervals[3].end.date() == date(2014, 12, 31)

    intervals = list(mixin.get_date_intervals(datetime(2016, 3, 1)))

    assert intervals[0].name == 'This month'
    assert intervals[0].start.date() == date(2016, 3, 1)
    assert intervals[0].end.date() == date(2016, 3, 31)

    assert intervals[1].name == 'Last month'
    assert intervals[1].start.date() == date(2016, 2, 1)
    assert intervals[1].end.date() == date(2016, 2, 29)

    assert intervals[2].name == 'This year'
    assert intervals[2].start.date() == date(2016, 1, 1)
    assert intervals[2].end.date() == date(2016, 1, 31)

    assert intervals[3].name == 'Last year'
    assert intervals[3].start.date() == date(2015, 1, 1)
    assert intervals[3].end.date() == date(2015, 12, 31)

    assert intervals[4].name == 'Older'
    assert intervals[4].start.date() == date(2000, 1, 1)
    assert intervals[4].end.date() == date(2014, 12, 31)


def test_image_grouping(session):

    collection = ImageFileCollection(session)

    def grouped_by_date(today):
        grouped = collection.grouped_by_date(today=today)
        return OrderedDict((g, [i[1] for i in items]) for g, items in grouped)

    def delete(images):
        for image in images:
            collection.delete(image)

    # catches all intervals
    images = [collection.add('x.png', create_image()) for r in range(0, 11)]
    images[0].created = datetime(2008, 9, 1, 16, 0, tzinfo=utc)   # next month
    images[1].created = datetime(2008, 8, 3, 16, 0, tzinfo=utc)   # this month
    images[2].created = datetime(2008, 8, 2, 16, 0, tzinfo=utc)   # this month
    images[3].created = datetime(2008, 8, 1, 0, 0, tzinfo=utc)    # this month
    images[4].created = datetime(2008, 7, 1, 16, 0, tzinfo=utc)   # last month
    images[5].created = datetime(2008, 5, 1, 16, 0, tzinfo=utc)   # this year
    images[6].created = datetime(2008, 2, 4, 12, 0, tzinfo=utc)   # this year
    images[7].created = datetime(2007, 12, 1, 16, 0, tzinfo=utc)  # last year
    images[8].created = datetime(2007, 10, 2, 14, 0, tzinfo=utc)  # last year
    images[9].created = datetime(2005, 4, 1, 16, 0, tzinfo=utc)   # older
    images[10].created = datetime(2002, 6, 4, 16, 0, tzinfo=utc)  # older

    grouped = grouped_by_date(datetime(2008, 8, 8))

    assert grouped['This month'] == [img.id for img in images[1:4]]
    assert grouped['Last month'] == [img.id for img in images[4:5]]
    assert grouped['This year'] == [img.id for img in images[5:7]]
    assert grouped['Last year'] == [img.id for img in images[7:9]]
    assert grouped['Older'] == [img.id for img in images[9:]]

    delete(images)

    # only catch some intervals
    images = [collection.add('x.png', create_image()) for r in range(0, 3)]
    images[0].created = datetime(2008, 8, 3, 16, 0, tzinfo=utc)  # this month
    images[1].created = datetime(2008, 5, 1, 16, 0, tzinfo=utc)  # this year
    images[2].created = datetime(2002, 6, 4, 16, 0, tzinfo=utc)  # older

    grouped = grouped_by_date(datetime(2008, 8, 8))
    assert grouped['This month'] == [img.id for img in images[0:1]]
    assert grouped['This year'] == [img.id for img in images[1:2]]
    assert grouped['Older'] == [img.id for img in images[2:]]

    delete(images)

    # catch a different set of intervals
    images = [collection.add('x.png', create_image()) for r in range(0, 2)]
    images[0].created = datetime(2008, 7, 1, 16, 0, tzinfo=utc)  # last month
    images[1].created = datetime(2002, 6, 4, 16, 0, tzinfo=utc)  # older

    grouped = grouped_by_date(datetime(2008, 8, 8))
    assert grouped['Last month'] == [img.id for img in images[0:1]]
    assert grouped['Older'] == [img.id for img in images[1:]]

    delete(images)

    # use a different date and catch a reduced set of intervals
    images = [collection.add('x.png', create_image()) for r in range(0, 7)]
    images[0].created = datetime(2008, 1, 3, 16, 0, tzinfo=utc)   # this month
    images[1].created = datetime(2008, 1, 2, 16, 0, tzinfo=utc)   # this month
    images[2].created = datetime(2008, 1, 1, 0, 0, tzinfo=utc)    # this month
    images[3].created = datetime(2007, 12, 14, 16, 0, tzinfo=utc)  # lst month
    images[4].created = datetime(2007, 12, 12, 16, 0, tzinfo=utc)  # lst month
    images[5].created = datetime(2007, 10, 2, 14, 0, tzinfo=utc)  # last year
    images[6].created = datetime(2002, 6, 4, 16, 0, tzinfo=utc)   # older

    grouped = grouped_by_date(datetime(2008, 1, 8))
    assert grouped['This month'] == [img.id for img in images[0:3]]
    assert grouped['Last month'] == [img.id for img in images[3:5]]
    assert grouped['Last year'] == [img.id for img in images[5:6]]
    assert grouped['Older'] == [img.id for img in images[6:]]

    delete(images)


def test_calendar_date_range():
    resource = SharedMethods()

    resource.date = None
    resource.timezone = utc

    resource.view = 'month'
    with freeze_time(datetime(2016, 5, 14, tzinfo=utc)):
        assert resource.calendar_date_range == (
            datetime(2016, 5, 1, tzinfo=utc),
            datetime(2016, 5, 31, 23, 59, 59, 999999, tzinfo=utc)
        )

    resource.date = date(2016, 5, 14)
    assert resource.calendar_date_range == (
        datetime(2016, 5, 1, tzinfo=utc),
        datetime(2016, 5, 31, 23, 59, 59, 999999, tzinfo=utc)
    )

    resource.view = 'agendaWeek'
    assert resource.calendar_date_range == (
        datetime(2016, 5, 9, tzinfo=utc),
        datetime(2016, 5, 15, 23, 59, 59, 999999, tzinfo=utc)
    )

    resource.view = 'agendaDay'
    assert resource.calendar_date_range == (
        datetime(2016, 5, 14, tzinfo=utc),
        datetime(2016, 5, 14, 23, 59, 59, 999999, tzinfo=utc)
    )


def test_sitecollection(org_app):

    sitecollection = SiteCollection(org_app.session())
    objects = sitecollection.get()

    assert {o.name for o in objects['topics']} == {
        'organisation',
        'themen',
        'kontakt',
    }

    assert {o.name for o in objects['news']} == {
        'news',
        'wir-haben-eine-neue-webseite'
    }

    builtin_forms_path = module_path('onegov.org', 'forms/builtin/de')

    paths = (p for p in os.listdir(builtin_forms_path))
    paths = (p for p in paths if p.endswith('.form'))
    paths = (os.path.basename(p) for p in paths)
    builtin_forms = set(rchop(p, '.form') for p in paths)

    assert {o.name for o in objects['forms']} == set(builtin_forms)



from onegov.form import Form
from onegov.org import _
from wtforms import BooleanField


class UserProfileForm(Form):
    """ Defines the settings form for user profiles. """

    daily_ticket_statistics = BooleanField(
        label=_("Send a daily status e-mail."),
        fieldset=_("General")
    )

    @property
    def enable_daily_ticket_statistics(self):
        roles = self.request.app.settings.org.status_mail_roles
        return self.request.current_role in roles

    def on_request(self):
        if not self.enable_daily_ticket_statistics:
            self.delete_field('daily_ticket_statistics')

    def populate_obj(self, model):
        super().populate_obj(model, exclude={
            'daily_ticket_statistics'
        })

        if self.enable_daily_ticket_statistics:
            model.data = model.data or {}
            model.data['daily_ticket_statistics']\
                = self.daily_ticket_statistics.data

    def process_obj(self, obj):
        super().process_obj(obj)

        if self.enable_daily_ticket_statistics:
            self.daily_ticket_statistics.data\
                = (obj.data or {}).get('daily_ticket_statistics', True)



from onegov.org.forms.allocation import DaypassAllocationEditForm
from onegov.org.forms.allocation import DaypassAllocationForm
from onegov.org.forms.allocation import RoomAllocationEditForm
from onegov.org.forms.allocation import RoomAllocationForm
from onegov.org.forms.directory import DirectoryForm
from onegov.org.forms.directory import DirectoryImportForm
from onegov.org.forms.event import EventForm
from onegov.org.forms.form_definition import FormDefinitionForm
from onegov.org.forms.form_export import FormSubmissionsExport
from onegov.org.forms.form_registration import FormRegistrationWindowForm
from onegov.org.forms.generic import ExportForm, DateRangeForm
from onegov.org.forms.imageset import ImageSetForm
from onegov.org.forms.newsletter import NewsletterForm
from onegov.org.forms.newsletter import NewsletterSendForm
from onegov.org.forms.newsletter import NewsletterTestForm
from onegov.org.forms.page import LinkForm, PageForm
from onegov.org.forms.person import PersonForm
from onegov.org.forms.reservation import ReservationForm
from onegov.org.forms.resource import ResourceCleanupForm
from onegov.org.forms.resource import ResourceExportForm
from onegov.org.forms.resource import ResourceForm
from onegov.org.forms.resource_recipient import ResourceRecipientForm
from onegov.org.forms.settings import AnalyticsSettingsForm
from onegov.org.forms.settings import FooterSettingsForm
from onegov.org.forms.settings import GeneralSettingsForm
from onegov.org.forms.settings import HomepageSettingsForm
from onegov.org.forms.settings import MapSettingsForm
from onegov.org.forms.settings import ModuleSettingsForm
from onegov.org.forms.signup import SignupForm
from onegov.org.forms.ticket_note import TicketNoteForm
from onegov.org.forms.user import ManageUserForm, NewUserForm
from onegov.org.forms.userprofile import UserProfileForm


__all__ = [
    'AnalyticsSettingsForm',
    'DateRangeForm',
    'DaypassAllocationEditForm',
    'DaypassAllocationForm',
    'DirectoryForm',
    'DirectoryImportForm',
    'EventForm',
    'ExportForm',
    'FooterSettingsForm',
    'FormDefinitionForm',
    'FormRegistrationWindowForm',
    'FormSubmissionsExport',
    'GeneralSettingsForm',
    'HomepageSettingsForm',
    'ImageSetForm',
    'LinkForm',
    'ManageUserForm',
    'MapSettingsForm',
    'ModuleSettingsForm',
    'NewsletterForm',
    'NewsletterSendForm',
    'NewsletterTestForm',
    'NewUserForm',
    'PageForm',
    'PersonForm',
    'ReservationForm',
    'ResourceCleanupForm',
    'ResourceExportForm',
    'ResourceForm',
    'ResourceRecipientForm',
    'RoomAllocationEditForm',
    'RoomAllocationForm',
    'SignupForm',
    'TicketNoteForm',
    'UserProfileForm',
]



from cached_property import cached_property


class Clipboard(object):
    """ The clipboard holds a url that should be copied and then pasted.
    The url is expected to be stored in a token that has been created by
    :meth:`onegov.core.request.CoreRequest.new_url_safe_token`.

    The reason behind this is that the url is used to fetch the object behind
    the url in an unrestricted fashion.

    """

    def __init__(self, request, token):
        self.request = request
        self.token = token

    @cached_property
    def url(self):
        data = self.request.load_url_safe_token(self.token, salt='clipboard')
        return data and data['url'] or None

    def get_object(self):
        if not self.url:
            return None

        return self.request.app.object_by_path(self.url)

    def clear(self):
        if self.request.browser_session.has('clipboard_url'):
            del self.request.browser_session['clipboard_url']

    def store_in_session(self):
        if self.url:
            self.request.browser_session['clipboard_url'] = self.url

    @classmethod
    def from_url(cls, request, url):
        return cls(
            request,
            request.new_url_safe_token({'url': url}, salt='clipboard')
        )

    @classmethod
    def from_session(cls, request):
        return cls(
            request,
            request.new_url_safe_token(
                {'url': request.browser_session.get('clipboard_url')},
                salt='clipboard'
            )
        )



import codecs
import os
import yaml

from datetime import datetime, timedelta
from onegov.core.cache import lru_cache
from onegov.core.utils import module_path
from onegov.event import EventCollection
from onegov.file import FileSetCollection, FileCollection
from onegov.form import FormCollection
from onegov.org import _
from onegov.org.models import Organisation
from onegov.page import PageCollection
from onegov.reservation import ResourceCollection
from pathlib import Path
from sedate import as_datetime


@lru_cache(maxsize=1)
def load_content(path):
    with open(path, 'r', encoding='utf-8') as f:
        return yaml.load(f)


def absolute_path(path, base):
    if path.startswith('/'):
        return path
    if path.startswith('~'):
        return os.path.expanduser(path)

    return os.path.join(base, path)


def create_new_organisation(app, name, create_files=True, path=None,
                            locale='de_CH'):
    locales = {
        'de_CH': 'content/de.yaml',
        'fr_CH': 'content/fr.yaml',
    }

    path = path or module_path('onegov.org', locales[locale])
    content = load_content(path)

    org = Organisation(name=name, **content['organisation'])
    org.meta['locales'] = locale

    session = app.session()
    session.add(org)

    translator = app.translations.get(locale)

    def translate(text):
        return text.interpolate(translator.gettext(text))

    add_pages(session, path)
    add_builtin_forms(session, locale=locale)
    add_events(session, name, translate, create_files)
    add_resources(app.libres_context, translate)

    if create_files:
        add_filesets(session, name, path)

    return org


def add_pages(session, path):
    pages = PageCollection(session)

    for ix, page in enumerate(load_content(path).get('pages')):
        if 'parent' in page:
            parent = pages.by_path(page['parent'])
        else:
            parent = None

        pages.add(
            parent=parent,
            title=page['title'],
            type=page['type'],
            name=page.get('name', None),
            meta=page.get('meta', None),
            content=page.get('content', None),
            order=ix
        )


def add_builtin_forms(session, definitions=None, locale='de_CH'):
    forms = FormCollection(session).definitions
    definitions = definitions or builtin_form_definitions(locale=locale)

    for name, title, definition in definitions:
        if not forms.by_name(name):
            form = forms.add(
                name=name,
                title=title,
                definition=definition,
                type='custom'
            )

        assert form.form_class().has_required_email_field, (
            "Each form must have at least one required email field"
        )


def builtin_form_definitions(path=None, locale='de_CH'):
    """ Yields the name, title and the form definition of all form definitions
    in the given or the default path.

    """
    locales = {
        'de_CH': 'forms/builtin/de',
        'fr_CH': 'forms/builtin/fr',
    }

    path = path or module_path('onegov.org', locales[locale])

    for filename in os.listdir(path):
        if filename.endswith('.form'):
            name = filename.replace('.form', '')
            title, definition = load_definition(os.path.join(path, filename))
            yield name, title, definition


def load_definition(path):
    """ Loads the title and the form definition from the given file. """

    with codecs.open(path, 'r', encoding='utf-8') as formfile:
        formlines = formfile.readlines()

        title = formlines[0].strip()
        definition = ''.join(formlines[3:])

        return title, definition


def add_resources(libres_context, translate):
    resource = ResourceCollection(libres_context)
    resource.add(
        translate(_("Daypass")),
        'Europe/Zurich',
        type='daypass',
        name='tageskarte'
    )
    resource.add(
        translate(_("Conference room")),
        'Europe/Zurich',
        type='room',
        name='konferenzraum'
    )


def add_filesets(session, organisation_name, path):
    base = os.path.dirname(path)

    for fileset in load_content(path).get('filesets', tuple()):

        fs = FileSetCollection(session, fileset['type']).add(
            title=fileset['title'],
            meta=fileset.get('meta', None),
            content=fileset.get('content', None)
        )

        files = FileCollection(session, fileset['type'])

        for file in fileset.get('files'):
            filepath = absolute_path(file['path'], base)

            with open(filepath, 'rb') as f:

                fs.files.append(
                    files.add(
                        filename=os.path.basename(file['path']),
                        content=f,
                        note=file.get('note', '').format(
                            organisation=organisation_name
                        )
                    )
                )


def add_events(session, name, translate, create_files):
    start = as_datetime(datetime.today().date())

    while start.weekday() != 6:
        start = start + timedelta(days=1)

    imgs = Path(module_path('onegov.org', 'content/images'))

    events = EventCollection(session)
    event = events.add(
        title=translate(_("150 years {organisation}")).format(
            organisation=name
        ),
        start=start + timedelta(hours=11, minutes=0),
        end=start + timedelta(hours=22, minutes=0),
        timezone="Europe/Zurich",
        tags=["Party"],
        location=translate(_("Sports facility")),
        content={
            "description": translate(_("We celebrate our 150th anniversary.")),
            "organizer": name
        },
        meta={"submitter_email": "info@example.org"}
    )

    if create_files:
        with (imgs / '3krgixyesbg-gregoire-bertaud.jpg').open('rb') as f:
            event.set_image(f)

    event.submit()
    event.publish()
    event = events.add(
        title=translate(_("General Assembly")),
        start=start + timedelta(days=2, hours=20, minutes=0),
        end=start + timedelta(days=2, hours=22, minutes=30),
        timezone="Europe/Zurich",
        tags=["Politics"],
        location=translate(_("Communal hall")),
        content={
            "description": translate(_("As every year.")),
            "organizer": name
        },
        meta={"submitter_email": "info@example.org"},
    )

    if create_files:
        with (imgs / '3avlwp-7bg8-mikael-kristenson.jpg').open('rb') as f:
            event.set_image(f)

    event.submit()
    event.publish()
    event = events.add(
        title=translate(_("Community Gymnastics")),
        start=start + timedelta(days=2, hours=10, minutes=0),
        end=start + timedelta(days=2, hours=11, minutes=0),
        recurrence=(
            "RRULE:FREQ=WEEKLY;WKST=MO;BYDAY=TU,TH;UNTIL={0}".format(
                (start + timedelta(days=31)).strftime('%Y%m%dT%H%M%SZ')
            )
        ),
        timezone="Europe/Zurich",
        tags=["Sports"],
        location=translate(_("Gymnasium")),
        content={
            "description": translate(_("Get fit together.")),
            "organizer": translate(_("Women's Club"))
        },
        meta={"submitter_email": "info@example.org"},
    )

    if create_files:
        with (imgs / 'uvdrvtwvqlu-swaminathan-jayaraman.jpg').open('rb') as f:
            event.set_image(f)

    event.submit()
    event.publish()
    event = events.add(
        title=translate(_("Football Tournament")),
        start=start + timedelta(days=7, hours=10, minutes=0),
        end=start + timedelta(days=7, hours=18, minutes=0),
        timezone="Europe/Zurich",
        tags=["Sports"],
        location=translate(_("Sports facility")),
        content={
            "description": translate(_("Amateurs welcome!")),
            "organizer": translate(_("Sports Association"))
        },
        meta={"submitter_email": "info@example.org"},
    )

    if create_files:
        with (imgs / 'eg4dcqkdoka-nathan-rogers.jpg').open('rb') as f:
            event.set_image(f)

    event.submit()
    event.publish()
